import sys
from pathlib import Path

# Add the repository root to PYTHONPATH
rtdl_path = Path.cwd()
while not (rtdl_path.name == 'rtdl' and rtdl_path.parent.name != 'rtdl'):
    rtdl_path = rtdl_path.parent
sys.path.append(str(rtdl_path))
import rtdl  # noqa

# >>> Project information <<<
author = 'rtdl authors'
copyright = '2021, rtdl authors'
project = 'rtdl'
release = rtdl.__version__
version = rtdl.__version__

# >>> General options <<<
default_role = 'py:obj'
pygments_style = 'default'
repo_url = 'https://github.com/Yura52/rtdl'
templates_path = ['_templates']

# >>> Extensions options <<<
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    # 'sphinxcontrib.spelling',
    'sphinx_copybutton',
]

autoclass_content = 'both'
autodoc_member_order = 'bysource'
autodoc_inherit_docstrings = False

doctest_global_setup = '''
import numpy as np
import torch
import rtdl
import rtdl.data
from rtdl import *
from rtdl.data import *
'''

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'torch': ('https://pytorch.org/docs/stable', None),
}

napoleon_numpy_docstring = False
napoleon_use_admonition_for_examples = False

# spelling_show_suggestions = True

# >>> HTML and theme options <<<
import sphinx_material  # noqa

html_static_path = ['_static']
html_theme = 'sphinx_material'
html_css_files = ['custom.css']
html_theme_options = {
    'base_url': 'https://Yura52.github.io/rtdl',
    'color_primary': 'red',
    'globaltoc_collapse': False,
    'globaltoc_depth': 2,
    'logo_icon': '&#127968;',
    'nav_links': [],
    'nav_title': project + ' ' + version,
    'repo_name': project,
    'repo_url': repo_url,
    'repo_type': 'github',
    'master_doc': False,
    'version_dropdown': True,
    'version_json': '_static/versions.json',
}
html_sidebars = {
    '**': ['logo-text.html', 'globaltoc.html', 'searchbox.html'],
}



import enum
import math
import time
import warnings
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim
from torch import Tensor

from . import functional as rtdlF

ModuleType = Union[str, Callable[..., nn.Module]]
_INTERNAL_ERROR_MESSAGE = 'Internal error. Please, open an issue.'


def _is_glu_activation(activation: ModuleType):
    return (
        isinstance(activation, str)
        and activation.endswith('GLU')
        or activation in [ReGLU, GEGLU]
    )


def _all_or_none(values):
    return all(x is None for x in values) or all(x is not None for x in values)


class ReGLU(nn.Module):
    """The ReGLU activation function from [shazeer2020glu].

    Examples:
        .. testcode::

            module = ReGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)

    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return rtdlF.reglu(x)


class GEGLU(nn.Module):
    """The GEGLU activation function from [shazeer2020glu].

    Examples:
        .. testcode::

            module = GEGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)

    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return rtdlF.geglu(x)


class _TokenInitialization(enum.Enum):
    UNIFORM = 'uniform'
    NORMAL = 'normal'

    @classmethod
    def from_str(cls, initialization: str) -> '_TokenInitialization':
        try:
            return cls(initialization)
        except ValueError:
            valid_values = [x.value for x in _TokenInitialization]
            raise ValueError(f'initialization must be one of {valid_values}')

    def apply(self, x: Tensor, d: int) -> None:
        d_sqrt_inv = 1 / math.sqrt(d)
        if self == _TokenInitialization.UNIFORM:
            # used in the paper "Revisiting Deep Learning Models for Tabular Data";
            # is equivalent to `nn.init.kaiming_uniform_(x, a=math.sqrt(5))` (which is
            # used by torch to initialize nn.Linear.weight, for example)
            nn.init.uniform_(x, a=-d_sqrt_inv, b=d_sqrt_inv)
        elif self == _TokenInitialization.NORMAL:
            nn.init.normal_(x, std=d_sqrt_inv)


class NumericalFeatureTokenizer(nn.Module):
    """Transforms continuous features to tokens (embeddings).

    See `FeatureTokenizer` for the illustration.

    For one feature, the transformation consists of two steps:

    * the feature is multiplied by a trainable vector
    * another trainable vector is added

    Note that each feature has its separate pair of trainable vectors, i.e. the vectors
    are not shared between features.

    Examples:
        .. testcode::

            x = torch.randn(4, 2)
            n_objects, n_features = x.shape
            d_token = 3
            tokenizer = NumericalFeatureTokenizer(n_features, d_token, True, 'uniform')
            tokens = tokenizer(x)
            assert tokens.shape == (n_objects, n_features, d_token)
    """

    def __init__(
        self,
        n_features: int,
        d_token: int,
        bias: bool,
        initialization: str,
    ) -> None:
        """
        Args:
            n_features: the number of continuous (scalar) features
            d_token: the size of one token
            bias: if `False`, then the transformation will include only multiplication.
                **Warning**: :code:`bias=False` leads to significantly worse results for
                Transformer-like (token-based) architectures.
            initialization: initialization policy for parameters. Must be one of
                :code:`['uniform', 'normal']`. Let :code:`s = d ** -0.5`. Then, the
                corresponding distributions are :code:`Uniform(-s, s)` and :code:`Normal(0, s)`.
                In [gorishniy2021revisiting], the 'uniform' initialization was used.

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        super().__init__()
        initialization_ = _TokenInitialization.from_str(initialization)
        self.weight = nn.Parameter(Tensor(n_features, d_token))
        self.bias = nn.Parameter(Tensor(n_features, d_token)) if bias else None
        for parameter in [self.weight, self.bias]:
            if parameter is not None:
                initialization_.apply(parameter, d_token)

    @property
    def n_tokens(self) -> int:
        """The number of tokens."""
        return len(self.weight)

    @property
    def d_token(self) -> int:
        """The size of one token."""
        return self.weight.shape[1]

    def forward(self, x: Tensor) -> Tensor:
        x = self.weight[None] * x[..., None]
        if self.bias is not None:
            x = x + self.bias[None]
        return x


class CategoricalFeatureTokenizer(nn.Module):
    """Transforms categorical features to tokens (embeddings).

    See `FeatureTokenizer` for the illustration.

    The module efficiently implements a collection of `torch.nn.Embedding` (with
    optional biases).

    Examples:
        .. testcode::

            # the input must contain integers. For example, if the first feature can
            # take 3 distinct values, then its cardinality is 3 and the first column
            # must contain values from the range `[0, 1, 2]`.
            cardinalities = [3, 10]
            x = torch.tensor([
                [0, 5],
                [1, 7],
                [0, 2],
                [2, 4]
            ])
            n_objects, n_features = x.shape
            d_token = 3
            tokenizer = CategoricalFeatureTokenizer(cardinalities, d_token, True, 'uniform')
            tokens = tokenizer(x)
            assert tokens.shape == (n_objects, n_features, d_token)
    """

    category_offsets: Tensor

    def __init__(
        self,
        cardinalities: List[int],
        d_token: int,
        bias: bool,
        initialization: str,
    ) -> None:
        """
        Args:
            cardinalities: the number of distinct values for each feature. For example,
                :code:`cardinalities=[3, 4]` describes two features: the first one can
                take values in the range :code:`[0, 1, 2]` and the second one can take
                values in the range :code:`[0, 1, 2, 3]`.
            d_token: the size of one token.
            bias: if `True`, for each feature, a trainable vector is added to the
                embedding regardless of feature value. The bias vectors are not shared
                between features.
            initialization: initialization policy for parameters. Must be one of
                :code:`['uniform', 'normal']`. Let :code:`s = d ** -0.5`. Then, the
                corresponding distributions are :code:`Uniform(-s, s)` and :code:`Normal(0, s)`. In
                the paper [gorishniy2021revisiting], the 'uniform' initialization was
                used.

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        super().__init__()
        assert cardinalities, 'cardinalities must be non-empty'
        assert d_token > 0, 'd_token must be positive'
        initialization_ = _TokenInitialization.from_str(initialization)

        category_offsets = torch.tensor([0] + cardinalities[:-1]).cumsum(0)
        self.register_buffer('category_offsets', category_offsets, persistent=False)
        self.embeddings = nn.Embedding(sum(cardinalities), d_token)
        self.bias = nn.Parameter(Tensor(len(cardinalities), d_token)) if bias else None

        for parameter in [self.embeddings.weight, self.bias]:
            if parameter is not None:
                initialization_.apply(parameter, d_token)

    @property
    def n_tokens(self) -> int:
        """The number of tokens."""
        return len(self.category_offsets)

    @property
    def d_token(self) -> int:
        """The size of one token."""
        return self.embeddings.embedding_dim

    def forward(self, x: Tensor) -> Tensor:
        x = self.embeddings(x + self.category_offsets[None])
        if self.bias is not None:
            x = x + self.bias[None]
        return x


class FeatureTokenizer(nn.Module):
    """Combines `NumericalFeatureTokenizer` and `CategoricalFeatureTokenizer`.

    The "Feature Tokenizer" module from [gorishniy2021revisiting]. The module transforms
    continuous and categorical features to tokens (embeddings).

    In the illustration below, the red module in the upper brackets represents
    `NumericalFeatureTokenizer` and the green module in the lower brackets represents
    `CategoricalFeatureTokenizer`.

    .. image:: ../images/feature_tokenizer.png
        :scale: 33%
        :alt: Feature Tokenizer

    Examples:
        .. testcode::

            n_objects = 4
            n_num_features = 3
            n_cat_features = 2
            d_token = 7
            x_num = torch.randn(n_objects, n_num_features)
            x_cat = torch.tensor([[0, 1], [1, 0], [0, 2], [1, 1]])
            # [2, 3] reflects cardinalities fr
            tokenizer = FeatureTokenizer(n_num_features, [2, 3], d_token)
            tokens = tokenizer(x_num, x_cat)
            assert tokens.shape == (n_objects, n_num_features + n_cat_features, d_token)

    References:
        * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko "Revisiting Deep Learning Models for Tabular Data", 2021
    """

    def __init__(
        self,
        n_num_features: int,
        cat_cardinalities: List[int],
        d_token: int,
    ) -> None:
        """
        Args:
            n_num_features: the number of continuous features. Pass :code:`0` if there
                are no numerical features.
            cat_cardinalities: the number of unique values for each feature. See
                `CategoricalFeatureTokenizer` for details. Pass an empty list if there
                are no categorical features.
            d_token: the size of one token.
        """
        super().__init__()
        assert n_num_features >= 0, 'n_num_features must be non-negative'
        assert (
            n_num_features or cat_cardinalities
        ), 'at least one of n_num_features or cat_cardinalities must be positive/non-empty'
        self.initialization = 'uniform'
        self.num_tokenizer = (
            NumericalFeatureTokenizer(
                n_features=n_num_features,
                d_token=d_token,
                bias=True,
                initialization=self.initialization,
            )
            if n_num_features
            else None
        )
        self.cat_tokenizer = (
            CategoricalFeatureTokenizer(
                cat_cardinalities, d_token, True, self.initialization
            )
            if cat_cardinalities
            else None
        )

    @property
    def n_tokens(self) -> int:
        """The number of tokens."""
        return sum(
            x.n_tokens
            for x in [self.num_tokenizer, self.cat_tokenizer]
            if x is not None
        )

    @property
    def d_token(self) -> int:
        """The size of one token."""
        return (
            self.cat_tokenizer.d_token  # type: ignore
            if self.num_tokenizer is None
            else self.num_tokenizer.d_token
        )

    def forward(self, x_num: Optional[Tensor], x_cat: Optional[Tensor]) -> Tensor:
        """Perform the forward pass.

        Args:
            x_num: continuous features. Must be presented if :code:`n_num_features > 0`
                was passed to the constructor.
            x_cat: categorical features (see `CategoricalFeatureTokenizer.forward` for
                details). Must be presented if non-empty :code:`cat_cardinalities` was
                passed to the constructor.
        Returns:
            tokens
        Raises:
            AssertionError: if the described requirements for the inputs are not met.
        """
        assert (
            x_num is not None or x_cat is not None
        ), 'At least one of x_num and x_cat must be presented'
        assert _all_or_none(
            [self.num_tokenizer, x_num]
        ), 'If self.num_tokenizer is (not) None, then x_num must (not) be None'
        assert _all_or_none(
            [self.cat_tokenizer, x_cat]
        ), 'If self.cat_tokenizer is (not) None, then x_cat must (not) be None'
        x = []
        if self.num_tokenizer is not None:
            x.append(self.num_tokenizer(x_num))
        if self.cat_tokenizer is not None:
            x.append(self.cat_tokenizer(x_cat))
        return x[0] if len(x) == 1 else torch.cat(x, dim=1)


class CLSToken(nn.Module):
    """[CLS]-token for BERT-like inference.

    To learn about the [CLS]-based inference, see [devlin2018bert].

    When used as a module, the [CLS]-token is appended **to the end** of each item in
    the batch.

    Examples:
        .. testcode::

            batch_size = 2
            n_tokens = 3
            d_token = 4
            cls_token = CLSToken(d_token, 'uniform')
            x = torch.randn(batch_size, n_tokens, d_token)
            x = cls_token(x)
            assert x.shape == (batch_size, n_tokens + 1, d_token)
            assert (x[:, -1, :] == cls_token.expand(len(x))).all()

    References:
        * [devlin2018bert] Jacob Devlin, Ming-Wei Chang, Kenton Lee, Kristina Toutanova "BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding" 2018
    """

    def __init__(self, d_token: int, initialization: str) -> None:
        """
        Args:
            d_token: the size of token
            initialization: initialization policy for parameters. Must be one of
                :code:`['uniform', 'normal']`. Let :code:`s = d ** -0.5`. Then, the
                corresponding distributions are :code:`Uniform(-s, s)` and :code:`Normal(0, s)`. In
                the paper [gorishniy2021revisiting], the 'uniform' initialization was
                used.

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        super().__init__()
        initialization_ = _TokenInitialization.from_str(initialization)
        self.weight = nn.Parameter(Tensor(d_token))
        initialization_.apply(self.weight, d_token)

    def expand(self, *leading_dimensions: int) -> Tensor:
        """Expand (repeat) the underlying [CLS]-token to a tensor with the given leading dimensions.

        A possible use case is building a batch of [CLS]-tokens. See `CLSToken` for
        examples of usage.

        Note:
            Under the hood, the `torch.Tensor.expand` method is applied to the
            underlying :code:`weight` parameter, so gradients will be propagated as
            expected.

        Args:
            leading_dimensions: the additional new dimensions

        Returns:
            tensor of the shape :code:`(*leading_dimensions, len(self.weight))`
        """
        if not leading_dimensions:
            return self.weight
        new_dims = (1,) * (len(leading_dimensions) - 1)
        return self.weight.view(*new_dims, -1).expand(*leading_dimensions, -1)

    def forward(self, x: Tensor) -> Tensor:
        """Append self **to the end** of each item in the batch (see `CLSToken`)."""
        return torch.cat([x, self.expand(len(x), 1)], dim=1)


def _make_nn_module(module_type: ModuleType, *args) -> nn.Module:
    if isinstance(module_type, str):
        if module_type == 'ReGLU':
            return ReGLU()
        elif module_type == 'GEGLU':
            return GEGLU()
        else:
            try:
                cls = getattr(nn, module_type)
            except AttributeError as err:
                raise ValueError(
                    f'Failed to construct the module {module_type} with the arguments {args}'
                ) from err
            return cls(*args)
    else:
        return module_type(*args)


class MLP(nn.Module):
    """The MLP model used in [gorishniy2021revisiting].

    The following scheme describes the architecture:

    .. code-block:: text

          MLP: (in) -> Block -> ... -> Block -> Linear -> (out)
        Block: (in) -> Linear -> Activation -> Dropout -> (out)

    Examples:
        .. testcode::

            x = torch.randn(4, 2)
            module = MLP.make_baseline(x.shape[1], [3, 5], 0.1, 1)
            assert module(x).shape == (len(x), 1)

    References:
        * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
    """

    class Block(nn.Module):
        """The main building block of `MLP`."""

        def __init__(
            self,
            *,
            d_in: int,
            d_out: int,
            bias: bool,
            activation: ModuleType,
            dropout: float,
        ) -> None:
            super().__init__()
            self.linear = nn.Linear(d_in, d_out, bias)
            self.activation = _make_nn_module(activation)
            self.dropout = nn.Dropout(dropout)

        def forward(self, x: Tensor) -> Tensor:
            return self.dropout(self.activation(self.linear(x)))

    def __init__(
        self,
        *,
        d_in: int,
        d_layers: List[int],
        dropouts: Union[float, List[float]],
        activation: Union[str, Callable[[], nn.Module]],
        d_out: int,
    ) -> None:
        """
        Note:
            `make_baseline` is the recommended constructor.
        """
        super().__init__()
        if isinstance(dropouts, float):
            dropouts = [dropouts] * len(d_layers)
        assert len(d_layers) == len(dropouts)

        self.blocks = nn.Sequential(
            *[
                MLP.Block(
                    d_in=d_layers[i - 1] if i else d_in,
                    d_out=d,
                    bias=True,
                    activation=activation,
                    dropout=dropout,
                )
                for i, (d, dropout) in enumerate(zip(d_layers, dropouts))
            ]
        )
        self.head = nn.Linear(d_layers[-1] if d_layers else d_in, d_out)

    @classmethod
    def make_baseline(
        cls: Type['MLP'],
        d_in: int,
        d_layers: List[int],
        dropout: float,
        d_out: int,
    ) -> 'MLP':
        """Create a "baseline" `MLP`.

        This variation of MLP was used in [gorishniy2021revisiting]. Features:

        * :code:`Activation` = :code:`ReLU`
        * all linear layers except for the first one and the last one are of the same dimension
        * the dropout rate is the same for all dropout layers

        Args:
            d_in: the input size
            d_layers: the dimensions of the linear layers. If there are more than two
                layers, then all of them except for the first and the last ones must
                have the same dimension. Valid examples: :code:`[]`, :code:`[8]`,
                :code:`[8, 16]`, :code:`[2, 2, 2, 2]`, :code:`[1, 2, 2, 4]`. Invalid
                example: :code:`[1, 2, 3, 4]`.
            dropout: the dropout rate for all hidden layers
            d_out: the output size
        Returns:
            MLP

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        assert isinstance(dropout, float), 'In this constructor, dropout must be float'
        if len(d_layers) > 2:
            assert len(set(d_layers[1:-1])) == 1, (
                'In this constructor, if d_layers contains more than two elements, then'
                ' all elements except for the first and the last ones must be equal.'
            )
        return MLP(
            d_in=d_in,
            d_layers=d_layers,  # type: ignore
            dropouts=dropout,
            activation='ReLU',
            d_out=d_out,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.blocks(x)
        x = self.head(x)
        return x


class ResNet(nn.Module):
    """The ResNet model used in [gorishniy2021revisiting].

    The following scheme describes the architecture:

    .. code-block:: text

        ResNet: (in) -> Linear -> Block -> ... -> Block -> Head -> (out)

                 |-> Norm -> Linear -> Activation -> Dropout -> Linear -> Dropout ->|
                 |                                                                  |
         Block: (in) ------------------------------------------------------------> Add -> (out)

          Head: (in) -> Norm -> Activation -> Linear -> (out)

    Examples:
        .. testcode::

            x = torch.randn(4, 2)
            module = ResNet.make_baseline(
                d_in=x.shape[1],
                n_blocks=2,
                d_main=3,
                d_hidden=4,
                dropout_first=0.25,
                dropout_second=0.0,
                d_out=1
            )
            assert module(x).shape == (len(x), 1)

    References:
        * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
    """

    class Block(nn.Module):
        """The main building block of `ResNet`."""

        def __init__(
            self,
            *,
            d_main: int,
            d_hidden: int,
            bias_first: bool,
            bias_second: bool,
            dropout_first: float,
            dropout_second: float,
            normalization: ModuleType,
            activation: ModuleType,
            skip_connection: bool,
        ) -> None:
            super().__init__()
            self.normalization = _make_nn_module(normalization, d_main)
            self.linear_first = nn.Linear(d_main, d_hidden, bias_first)
            self.activation = _make_nn_module(activation)
            self.dropout_first = nn.Dropout(dropout_first)
            self.linear_second = nn.Linear(d_hidden, d_main, bias_second)
            self.dropout_second = nn.Dropout(dropout_second)
            self.skip_connection = skip_connection

        def forward(self, x: Tensor) -> Tensor:
            x_input = x
            x = self.normalization(x)
            x = self.linear_first(x)
            x = self.activation(x)
            x = self.dropout_first(x)
            x = self.linear_second(x)
            x = self.dropout_second(x)
            if self.skip_connection:
                x = x_input + x
            return x

    class Head(nn.Module):
        """The final module of `ResNet`."""

        def __init__(
            self,
            *,
            d_in: int,
            d_out: int,
            bias: bool,
            normalization: ModuleType,
            activation: ModuleType,
        ) -> None:
            super().__init__()
            self.normalization = _make_nn_module(normalization, d_in)
            self.activation = _make_nn_module(activation)
            self.linear = nn.Linear(d_in, d_out, bias)

        def forward(self, x: Tensor) -> Tensor:
            if self.normalization is not None:
                x = self.normalization(x)
            x = self.activation(x)
            x = self.linear(x)
            return x

    def __init__(
        self,
        *,
        d_in: int,
        n_blocks: int,
        d_main: int,
        d_hidden: int,
        dropout_first: float,
        dropout_second: float,
        normalization: ModuleType,
        activation: ModuleType,
        d_out: int,
    ) -> None:
        """
        Note:
            `make_baseline` is the recommended constructor.
        """
        super().__init__()

        self.first_layer = nn.Linear(d_in, d_main)
        if d_main is None:
            d_main = d_in
        self.blocks = nn.Sequential(
            *[
                ResNet.Block(
                    d_main=d_main,
                    d_hidden=d_hidden,
                    bias_first=True,
                    bias_second=True,
                    dropout_first=dropout_first,
                    dropout_second=dropout_second,
                    normalization=normalization,
                    activation=activation,
                    skip_connection=True,
                )
                for _ in range(n_blocks)
            ]
        )
        self.head = ResNet.Head(
            d_in=d_main,
            d_out=d_out,
            bias=True,
            normalization=normalization,
            activation=activation,
        )

    @classmethod
    def make_baseline(
        cls: Type['ResNet'],
        *,
        d_in: int,
        n_blocks: int,
        d_main: int,
        d_hidden: int,
        dropout_first: float,
        dropout_second: float,
        d_out: int,
    ) -> 'ResNet':
        """Create a "baseline" `ResNet`.

        This variation of ResNet was used in [gorishniy2021revisiting]. Features:

        * :code:`Activation` = :code:`ReLU`
        * :code:`Norm` = :code:`BatchNorm1d`

        Args:
            d_in: the input size
            n_blocks: the number of Blocks
            d_main: the input size (or, equivalently, the output size) of each Block
            d_hidden: the output size of the first linear layer in each Block
            dropout_first: the dropout rate of the first dropout layer in each Block.
            dropout_second: the dropout rate of the second dropout layer in each Block.

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        return cls(
            d_in=d_in,
            n_blocks=n_blocks,
            d_main=d_main,
            d_hidden=d_hidden,
            dropout_first=dropout_first,
            dropout_second=dropout_second,
            normalization='BatchNorm1d',
            activation='ReLU',
            d_out=d_out,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.first_layer(x)
        x = self.blocks(x)
        x = self.head(x)
        return x


class MultiheadAttention(nn.Module):
    """Multihead Attention (self-/cross-) with optional 'linear' attention.

    To learn more about Multihead Attention, see [devlin2018bert]. See the implementation
    of `Transformer` and the examples below to learn how to use the compression technique
    from [wang2020linformer] to speed up the module when the number of tokens is large.

    Examples:
        .. testcode::

            n_objects, n_tokens, d_token = 2, 3, 12
            n_heads = 6
            a = torch.randn(n_objects, n_tokens, d_token)
            b = torch.randn(n_objects, n_tokens * 2, d_token)
            module = MultiheadAttention(
                d_token=d_token, n_heads=n_heads, dropout=0.2, bias=True, initialization='kaiming'
            )

            # self-attention
            x, attention_stats = module(a, a, None, None)
            assert x.shape == a.shape
            assert attention_stats['attention_probs'].shape == (n_objects * n_heads, n_tokens, n_tokens)
            assert attention_stats['attention_logits'].shape == (n_objects * n_heads, n_tokens, n_tokens)

            # cross-attention
            assert module(a, b, None, None)

            # Linformer self-attention with the 'headwise' sharing policy
            k_compression = torch.nn.Linear(n_tokens, n_tokens // 4)
            v_compression = torch.nn.Linear(n_tokens, n_tokens // 4)
            assert module(a, a, k_compression, v_compression)

            # Linformer self-attention with the 'key-value' sharing policy
            kv_compression = torch.nn.Linear(n_tokens, n_tokens // 4)
            assert module(a, a, kv_compression, kv_compression)

    References:
        * [devlin2018bert] Jacob Devlin, Ming-Wei Chang, Kenton Lee, Kristina Toutanova "BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding" 2018
        * [wang2020linformer] Sinong Wang, Belinda Z. Li, Madian Khabsa, Han Fang, Hao Ma "Linformer: Self-Attention with Linear Complexity", 2020
    """

    def __init__(
        self,
        *,
        d_token: int,
        n_heads: int,
        dropout: float,
        bias: bool,
        initialization: str,
    ) -> None:
        """
        Args:
            d_token: the token size. Must be a multiple of :code:`n_heads`.
            n_heads: the number of heads. If greater than 1, then the module will have
                an addition output layer (so called "mixing" layer).
            dropout: dropout rate for the attention map. The dropout is applied to
                *probabilities* and do not affect logits.
            bias: if `True`, then input (and output, if presented) layers also have bias.
                `True` is a reasonable default choice.
            initialization: initialization for input projection layers. Must be one of
                :code:`['kaiming', 'xavier']`. `kaiming` is a reasonable default choice.
        Raises:
            AssertionError: if requirements for the inputs are not met.
        """
        super().__init__()
        if n_heads > 1:
            assert d_token % n_heads == 0, 'd_token must be a multiple of n_heads'
        assert initialization in ['kaiming', 'xavier']

        self.W_q = nn.Linear(d_token, d_token, bias)
        self.W_k = nn.Linear(d_token, d_token, bias)
        self.W_v = nn.Linear(d_token, d_token, bias)
        self.W_out = nn.Linear(d_token, d_token, bias) if n_heads > 1 else None
        self.n_heads = n_heads
        self.dropout = nn.Dropout(dropout) if dropout else None

        for m in [self.W_q, self.W_k, self.W_v]:
            # the "xavier" branch tries to follow torch.nn.MultiheadAttention;
            # the second condition checks if W_v plays the role of W_out; the latter one
            # is initialized with Kaiming in torch
            if initialization == 'xavier' and (
                m is not self.W_v or self.W_out is not None
            ):
                # gain is needed since W_qkv is represented with 3 separate layers (it
                # implies different fan_out)
                nn.init.xavier_uniform_(m.weight, gain=1 / math.sqrt(2))
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        if self.W_out is not None:
            nn.init.zeros_(self.W_out.bias)

    def _reshape(self, x: Tensor) -> Tensor:
        batch_size, n_tokens, d = x.shape
        d_head = d // self.n_heads
        return (
            x.reshape(batch_size, n_tokens, self.n_heads, d_head)
            .transpose(1, 2)
            .reshape(batch_size * self.n_heads, n_tokens, d_head)
        )

    def forward(
        self,
        x_q: Tensor,
        x_kv: Tensor,
        key_compression: Optional[nn.Linear],
        value_compression: Optional[nn.Linear],
    ) -> Tuple[Tensor, Dict[str, Tensor]]:
        """Perform the forward pass.

        Args:
            x_q: query tokens
            x_kv: key-value tokens
            key_compression: Linformer-style compression for keys
            value_compression: Linformer-style compression for values
        Returns:
            (tokens, attention_stats)
        """
        assert _all_or_none(
            [key_compression, value_compression]
        ), 'If key_compression is (not) None, then value_compression must (not) be None'
        q, k, v = self.W_q(x_q), self.W_k(x_kv), self.W_v(x_kv)
        for tensor in [q, k, v]:
            assert tensor.shape[-1] % self.n_heads == 0, _INTERNAL_ERROR_MESSAGE
        if key_compression is not None:
            k = key_compression(k.transpose(1, 2)).transpose(1, 2)
            v = value_compression(v.transpose(1, 2)).transpose(1, 2)  # type: ignore

        batch_size = len(q)
        d_head_key = k.shape[-1] // self.n_heads
        d_head_value = v.shape[-1] // self.n_heads
        n_q_tokens = q.shape[1]

        q = self._reshape(q)
        k = self._reshape(k)
        attention_logits = q @ k.transpose(1, 2) / math.sqrt(d_head_key)
        attention_probs = F.softmax(attention_logits, dim=-1)
        if self.dropout is not None:
            attention_probs = self.dropout(attention_probs)
        x = attention_probs @ self._reshape(v)
        x = (
            x.reshape(batch_size, self.n_heads, n_q_tokens, d_head_value)
            .transpose(1, 2)
            .reshape(batch_size, n_q_tokens, self.n_heads * d_head_value)
        )
        if self.W_out is not None:
            x = self.W_out(x)
        return x, {
            'attention_logits': attention_logits,
            'attention_probs': attention_probs,
        }


class Transformer(nn.Module):
    """Transformer with extra features.

    This module is the backbone of `FTTransformer`."""

    WARNINGS = {'first_prenormalization': True, 'prenormalization': True}

    class FFN(nn.Module):
        """The Feed-Forward Network module used in every `Transformer` block."""

        def __init__(
            self,
            *,
            d_token: int,
            d_hidden: int,
            bias_first: bool,
            bias_second: bool,
            dropout: float,
            activation: ModuleType,
        ):
            super().__init__()
            self.linear_first = nn.Linear(
                d_token,
                d_hidden * (2 if _is_glu_activation(activation) else 1),
                bias_first,
            )
            self.activation = _make_nn_module(activation)
            self.dropout = nn.Dropout(dropout)
            self.linear_second = nn.Linear(d_hidden, d_token, bias_second)

        def forward(self, x: Tensor) -> Tensor:
            x = self.linear_first(x)
            x = self.activation(x)
            x = self.dropout(x)
            x = self.linear_second(x)
            return x

    class Head(nn.Module):
        """The final module of the `Transformer` that performs BERT-like inference."""

        def __init__(
            self,
            *,
            d_in: int,
            bias: bool,
            activation: ModuleType,
            normalization: ModuleType,
            d_out: int,
        ):
            super().__init__()
            self.normalization = _make_nn_module(normalization, d_in)
            self.activation = _make_nn_module(activation)
            self.linear = nn.Linear(d_in, d_out, bias)

        def forward(self, x: Tensor) -> Tensor:
            x = x[:, -1]
            x = self.normalization(x)
            x = self.activation(x)
            x = self.linear(x)
            return x

    def __init__(
        self,
        *,
        d_token: int,
        n_blocks: int,
        attention_n_heads: int,
        attention_dropout: float,
        attention_initialization: str,
        attention_normalization: str,
        ffn_d_hidden: int,
        ffn_dropout: float,
        ffn_activation: str,
        ffn_normalization: str,
        residual_dropout: float,
        prenormalization: bool,
        first_prenormalization: bool,
        last_layer_query_idx: Union[None, List[int], slice],
        n_tokens: Optional[int],
        kv_compression_ratio: Optional[float],
        kv_compression_sharing: Optional[str],
        head_activation: ModuleType,
        head_normalization: ModuleType,
        d_out: int,
    ) -> None:
        super().__init__()
        if isinstance(last_layer_query_idx, int):
            raise ValueError(
                'last_layer_query_idx must be None, list[int] or slice. '
                f'Do you mean last_layer_query_idx=[{last_layer_query_idx}] ?'
            )
        if not prenormalization:
            assert (
                not first_prenormalization
            ), 'If `prenormalization` is False, then `first_prenormalization` must be False'
        assert _all_or_none([n_tokens, kv_compression_ratio, kv_compression_sharing]), (
            'If any of the following arguments is (not) None, then all of them must (not) be None: '
            'n_tokens, kv_compression_ratio, kv_compression_sharing'
        )
        assert kv_compression_sharing in [None, 'headwise', 'key-value', 'layerwise']
        if not prenormalization:
            if self.WARNINGS['prenormalization']:
                warnings.warn(
                    'prenormalization is set to False. Are you sure about this? '
                    'The training can become less stable. '
                    'You can turn off this warning by tweaking the '
                    'rtdl.Transformer.WARNINGS dictionary.',
                    UserWarning,
                )
            assert (
                not first_prenormalization
            ), 'If prenormalization is False, then first_prenormalization is ignored and must be set to False'
        if (
            prenormalization
            and first_prenormalization
            and self.WARNINGS['first_prenormalization']
        ):
            warnings.warn(
                'first_prenormalization is set to True. Are you sure about this? '
                'For example, the vanilla FTTransformer with '
                'first_prenormalization=True performs SIGNIFICANTLY worse. '
                'You can turn off this warning by tweaking the '
                'rtdl.Transformer.WARNINGS dictionary.',
                UserWarning,
            )
            time.sleep(3)

        def make_kv_compression():
            assert (
                n_tokens and kv_compression_ratio
            ), _INTERNAL_ERROR_MESSAGE  # for mypy
            # https://github.com/pytorch/fairseq/blob/1bba712622b8ae4efb3eb793a8a40da386fe11d0/examples/linformer/linformer_src/modules/multihead_linear_attention.py#L83
            return nn.Linear(n_tokens, int(n_tokens * kv_compression_ratio), bias=False)

        self.shared_kv_compression = (
            make_kv_compression()
            if kv_compression_ratio and kv_compression_sharing == 'layerwise'
            else None
        )

        self.prenormalization = prenormalization
        self.last_layer_query_idx = last_layer_query_idx

        self.blocks = nn.ModuleList([])
        for layer_idx in range(n_blocks):
            layer = nn.ModuleDict(
                {
                    'attention': MultiheadAttention(
                        d_token=d_token,
                        n_heads=attention_n_heads,
                        dropout=attention_dropout,
                        bias=True,
                        initialization=attention_initialization,
                    ),
                    'ffn': Transformer.FFN(
                        d_token=d_token,
                        d_hidden=ffn_d_hidden,
                        bias_first=True,
                        bias_second=True,
                        dropout=ffn_dropout,
                        activation=ffn_activation,
                    ),
                    'attention_residual_dropout': nn.Dropout(residual_dropout),
                    'ffn_residual_dropout': nn.Dropout(residual_dropout),
                    'output': nn.Identity(),  # for hooks-based introspection
                }
            )
            if layer_idx or not prenormalization or first_prenormalization:
                layer['attention_normalization'] = _make_nn_module(
                    attention_normalization, d_token
                )
            layer['ffn_normalization'] = _make_nn_module(ffn_normalization, d_token)
            if kv_compression_ratio and self.shared_kv_compression is None:
                layer['key_compression'] = make_kv_compression()
                if kv_compression_sharing == 'headwise':
                    layer['value_compression'] = make_kv_compression()
                else:
                    assert (
                        kv_compression_sharing == 'key-value'
                    ), _INTERNAL_ERROR_MESSAGE
            self.blocks.append(layer)

        self.head = Transformer.Head(
            d_in=d_token,
            d_out=d_out,
            bias=True,
            activation=head_activation,  # type: ignore
            normalization=head_normalization if prenormalization else 'Identity',
        )

    def _get_kv_compressions(self, layer):
        return (
            (self.shared_kv_compression, self.shared_kv_compression)
            if self.shared_kv_compression is not None
            else (layer['key_compression'], layer['value_compression'])
            if 'key_compression' in layer and 'value_compression' in layer
            else (layer['key_compression'], layer['key_compression'])
            if 'key_compression' in layer
            else (None, None)
        )

    def _start_residual(self, layer, stage, x):
        assert stage in ['attention', 'ffn'], _INTERNAL_ERROR_MESSAGE
        x_residual = x
        if self.prenormalization:
            norm_key = f'{stage}_normalization'
            if norm_key in layer:
                x_residual = layer[norm_key](x_residual)
        return x_residual

    def _end_residual(self, layer, stage, x, x_residual):
        assert stage in ['attention', 'ffn'], _INTERNAL_ERROR_MESSAGE
        x_residual = layer[f'{stage}_residual_dropout'](x_residual)
        x = x + x_residual
        if not self.prenormalization:
            x = layer[f'{stage}_normalization'](x)
        return x

    def forward(self, x: Tensor) -> Tensor:
        assert (
            x.ndim == 3
        ), 'The input must have 3 dimensions: (n_objects, n_tokens, d_token)'
        for layer_idx, layer in enumerate(self.blocks):
            layer = cast(nn.ModuleDict, layer)

            query_idx = (
                self.last_layer_query_idx if layer_idx + 1 == len(self.blocks) else None
            )
            x_residual = self._start_residual(layer, 'attention', x)
            x_residual, _ = layer['attention'](
                x_residual if query_idx is None else x_residual[:, query_idx],
                x_residual,
                *self._get_kv_compressions(layer),
            )
            if query_idx is not None:
                x = x[:, query_idx]
            x = self._end_residual(layer, 'attention', x, x_residual)

            x_residual = self._start_residual(layer, 'ffn', x)
            x_residual = layer['ffn'](x_residual)
            x = self._end_residual(layer, 'ffn', x, x_residual)
            x = layer['output'](x)

        x = self.head(x)
        return x


class FTTransformer(nn.Module):
    """The FT-Transformer model proposed in [gorishniy2021revisiting].

    Transforms features to tokens with `FeatureTokenizer` and applies `Transformer` [vaswani2017attention]
    to the tokens. The following illustration provides a high-level overview of the
    architecture:

    .. image:: ../images/ft_transformer.png
        :scale: 25%
        :alt: FT-Transformer

    The following illustration demonstrates one Transformer block for :code:`prenormalization=True`:

    .. image:: ../images/transformer_block.png
        :scale: 25%
        :alt: PreNorm Transformer block

    Examples:
        .. testcode::

            x_num = torch.randn(4, 3)
            x_cat = torch.tensor([[0, 1], [1, 0], [0, 2], [1, 1]])

            module = FTTransformer.make_baseline(
                n_num_features=3,
                cat_cardinalities=[2, 3],
                d_token=8,
                n_blocks=2,
                attention_dropout=0.2,
                ffn_d_hidden=6,
                ffn_dropout=0.2,
                residual_dropout=0.0,
                d_out=1,
            )
            x = module(x_num, x_cat)
            assert x.shape == (4, 1)

            module = FTTransformer.make_default(
                n_num_features=3,
                cat_cardinalities=[2, 3],
                d_out=1,
            )
            x = module(x_num, x_cat)
            assert x.shape == (4, 1)

        To learn more about the baseline and default parameters:

        .. testcode::

            baseline_parameters = FTTransformer.get_baseline_transformer_subconfig()
            default_parameters = FTTransformer.get_default_transformer_config()

    References:
        * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        * [vaswani2017attention] Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N. Gomez, Lukasz Kaiser, Illia Polosukhin, "Attention Is All You Need", 2017
    """

    def __init__(
        self, feature_tokenizer: FeatureTokenizer, transformer: Transformer
    ) -> None:
        """
        Note:
            `make_baseline` and `make_default` are the recommended constructors.
        """
        super().__init__()
        if transformer.prenormalization:
            assert 'attention_normalization' not in transformer.blocks[0], (
                'In the prenormalization setting, FT-Transformer does not '
                'allow using the first normalization layer '
                'in the first transformer block'
            )
        self.feature_tokenizer = feature_tokenizer
        self.cls_token = CLSToken(
            feature_tokenizer.d_token, feature_tokenizer.initialization
        )
        self.transformer = transformer

    @classmethod
    def get_baseline_transformer_subconfig(
        cls: Type['FTTransformer'],
    ) -> Dict[str, Any]:
        """Get the baseline subset of parameters for the backbone."""
        return {
            'attention_n_heads': 8,
            'attention_initialization': 'kaiming',
            'ffn_activation': 'ReGLU',
            'attention_normalization': 'LayerNorm',
            'ffn_normalization': 'LayerNorm',
            'prenormalization': True,
            'first_prenormalization': False,
            'last_layer_query_idx': None,
            'n_tokens': None,
            'kv_compression_ratio': None,
            'kv_compression_sharing': None,
            'head_activation': 'ReLU',
            'head_normalization': 'LayerNorm',
        }

    @classmethod
    def get_default_transformer_config(
        cls: Type['FTTransformer'], *, n_blocks: int = 3
    ) -> Dict[str, Any]:
        """Get the default parameters for the backbone.

        Note:
            The configurations are different for different values of:code:`n_blocks`.
        """
        assert 1 <= n_blocks <= 6
        grid = {
            'd_token': [96, 128, 192, 256, 320, 384],
            'attention_dropout': [0.1, 0.15, 0.2, 0.25, 0.3, 0.35],
            'ffn_dropout': [0.0, 0.05, 0.1, 0.15, 0.2, 0.25],
        }
        arch_subconfig = {k: v[n_blocks - 1] for k, v in grid.items()}  # type: ignore
        baseline_subconfig = cls.get_baseline_transformer_subconfig()
        # (4 / 3) for ReGLU/GEGLU activations results in almost the same parameter count
        # as (2.0) for element-wise activations (e.g. ReLU or GELU; see the "else" branch)
        ffn_d_hidden_factor = (
            (4 / 3) if _is_glu_activation(baseline_subconfig['ffn_activation']) else 2.0
        )
        return {
            'n_blocks': n_blocks,
            'residual_dropout': 0.0,
            'ffn_d_hidden': int(arch_subconfig['d_token'] * ffn_d_hidden_factor),
            **arch_subconfig,
            **baseline_subconfig,
        }

    @classmethod
    def _make(
        cls,
        n_num_features,
        cat_cardinalities,
        transformer_config,
    ):
        feature_tokenizer = FeatureTokenizer(
            n_num_features=n_num_features,
            cat_cardinalities=cat_cardinalities,
            d_token=transformer_config['d_token'],
        )
        if transformer_config['d_out'] is None:
            transformer_config['head_activation'] = None
        if transformer_config['kv_compression_ratio'] is not None:
            transformer_config['n_tokens'] = feature_tokenizer.n_tokens + 1
        return FTTransformer(
            feature_tokenizer,
            Transformer(**transformer_config),
        )

    @classmethod
    def make_baseline(
        cls: Type['FTTransformer'],
        *,
        n_num_features: int,
        cat_cardinalities: Optional[List[int]],
        d_token: int,
        n_blocks: int,
        attention_dropout: float,
        ffn_d_hidden: int,
        ffn_dropout: float,
        residual_dropout: float,
        last_layer_query_idx: Union[None, List[int], slice] = None,
        kv_compression_ratio: Optional[float] = None,
        kv_compression_sharing: Optional[str] = None,
        d_out: int,
    ) -> 'FTTransformer':
        """Create a "baseline" `FTTransformer`.

        This variation of FT-Transformer was used in [gorishniy2021revisiting]. See
        `get_baseline_transformer_subconfig` to learn the values of other parameters.
        See `FTTransformer` for usage examples.

        Tip:
            `get_default_transformer_config` can serve as a starting point for choosing
            hyperparameter values.

        Args:
            n_num_features: the number of continuous features
            cat_cardinalities: the cardinalities of categorical features (see
                `CategoricalFeatureTokenizer` to learn more about cardinalities)
            d_token: the token size for each feature. Must be a multiple of :code:`n_heads=8`.
            n_blocks: the number of Transformer blocks
            attention_dropout: the dropout for attention blocks (see `MultiheadAttention`).
                Usually, positive values work better (even when the number of features is low).
            ffn_d_hidden: the *input* size for the *second* linear layer in `Transformer.FFN`.
                Note that it can be different from the output size of the first linear
                layer, since activations such as ReGLU or GEGLU change the size of input.
                For example, if :code:`ffn_d_hidden=10` and the activation is ReGLU (which
                is always true for the baseline and default configurations), then the
                output size of the first linear layer will be set to :code:`20`.
            ffn_dropout: the dropout rate after the first linear layer in `Transformer.FFN`.
            residual_dropout: the dropout rate for the output of each residual branch of
                all Transformer blocks.
            last_layer_query_idx: indices of tokens that should be processed by the last
                Transformer block. Note that for most cases there is no need to apply
                the last Transformer block to anything except for the [CLS]-token. Hence,
                runtime and memory can be saved by setting :code:`last_layer_query_idx=[-1]`,
                since the :code:`-1` is the position of [CLS]-token in FT-Transformer.
                Note that this will not affect the result in any way.
            kv_compression_ratio: apply the technique from [wang2020linformer] to speed
                up attention modules when the number of features is large. Can actually
                slow things down if the number of features is too low. Note that this
                option can affect task metrics in unpredictable way. Overall, use this
                option with caution. See `MultiheadAttention` for some examples and the
                implementation of `Transformer` to see how this option is used.
            kv_compression_sharing: weight sharing policy for :code:`kv_compression_ratio`.
                Must be one of :code:`[None, 'headwise', 'key-value', 'layerwise']`.
                See [wang2020linformer] to learn more about sharing policies.
                :code:`headwise` and :code:`key-value` are reasonable default choices. If
                :code:`kv_compression_ratio` is `None`, then this parameter also must be
                `None`. Otherwise, it must not be `None` (compression parameters must be
                shared in some way).
            d_out: the output size.

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
            * [wang2020linformer] Sinong Wang, Belinda Z. Li, Madian Khabsa, Han Fang, Hao Ma "Linformer: Self-Attention with Linear Complexity", 2020
        """
        transformer_config = cls.get_baseline_transformer_subconfig()
        for arg_name in [
            'n_blocks',
            'd_token',
            'attention_dropout',
            'ffn_d_hidden',
            'ffn_dropout',
            'residual_dropout',
            'last_layer_query_idx',
            'kv_compression_ratio',
            'kv_compression_sharing',
            'd_out',
        ]:
            transformer_config[arg_name] = locals()[arg_name]
        return cls._make(n_num_features, cat_cardinalities, transformer_config)

    @classmethod
    def make_default(
        cls: Type['FTTransformer'],
        *,
        n_num_features: int,
        cat_cardinalities: Optional[List[int]],
        n_blocks: int = 3,
        last_layer_query_idx: Union[None, List[int], slice] = None,
        kv_compression_ratio: Optional[float] = None,
        kv_compression_sharing: Optional[str] = None,
        d_out: int,
    ) -> 'FTTransformer':
        """Create the default `FTTransformer`.

        With :code:`n_blocks=3` (default) it is the FT-Transformer variation that is
        referred to as "default FT-Transformer" in [gorishniy2021revisiting]. See
        `FTTransformer` for usage examples. See `FTTransformer.make_baseline` for
        parameter descriptions.

        Note:
            The second component of the default FT-Transformer is the default optimizer,
            which can be created with the `make_default_optimizer` method.

        Note:
            According to [gorishniy2021revisiting], the default FT-Transformer is
            effective in the ensembling mode (i.e. when predictions of several default
            FT-Transformers are averaged). For a single FT-Transformer, it is still
            possible to achieve better results by tuning hyperparameter for the
            `make_baseline` constructor.

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        transformer_config = cls.get_default_transformer_config(n_blocks=n_blocks)
        for arg_name in [
            'last_layer_query_idx',
            'kv_compression_ratio',
            'kv_compression_sharing',
            'd_out',
        ]:
            transformer_config[arg_name] = locals()[arg_name]
        return cls._make(n_num_features, cat_cardinalities, transformer_config)

    def optimization_param_groups(self) -> List[Dict[str, Any]]:
        """The replacement for :code:`.parameters()` when creating optimizers.

        Example::

            optimizer = AdamW(
                model.optimization_param_groups(), lr=1e-4, weight_decay=1e-5
            )
        """
        no_wd_names = ['feature_tokenizer', 'normalization', '.bias']
        assert isinstance(
            getattr(self, no_wd_names[0], None), FeatureTokenizer
        ), _INTERNAL_ERROR_MESSAGE
        assert (
            sum(1 for name, _ in self.named_modules() if no_wd_names[1] in name)
            == len(self.transformer.blocks) * 2
            - int('attention_normalization' not in self.transformer.blocks[0])  # type: ignore
            + 1
        ), _INTERNAL_ERROR_MESSAGE

        def needs_wd(name):
            return all(x not in name for x in no_wd_names)

        return [
            {'params': [v for k, v in self.named_parameters() if needs_wd(k)]},
            {
                'params': [v for k, v in self.named_parameters() if not needs_wd(k)],
                'weight_decay': 0.0,
            },
        ]

    def make_default_optimizer(self) -> torch.optim.AdamW:
        """Make the optimizer for the default FT-Transformer."""
        return torch.optim.AdamW(
            self.optimization_param_groups(),
            lr=1e-4,
            weight_decay=1e-5,
        )

    def forward(self, x_num: Optional[Tensor], x_cat: Optional[Tensor]) -> Tensor:
        x = self.feature_tokenizer(x_num, x_cat)
        x = self.cls_token(x)
        x = self.transformer(x)
        return x



import pytest
import torch

import rtdl


def get_devices():
    return ['cpu', 'cuda'] if torch.cuda.is_available() else ['cpu']


def test_bad_mlp():
    with pytest.raises(AssertionError):
        rtdl.MLP.make_baseline(1, [1, 2, 3, 4], 0.0, 1)


@pytest.mark.parametrize('n_blocks', range(5))
@pytest.mark.parametrize('d_out', [1, 2])
@pytest.mark.parametrize('constructor', range(2))
@pytest.mark.parametrize('device', get_devices())
def test_mlp(n_blocks, d_out, constructor, device):
    if not n_blocks and not d_out:
        return
    d = 4
    d_last = d + 1
    d_layers = []
    if n_blocks:
        d_layers.append(d)
    if n_blocks > 2:
        d_layers.extend([d + d_out] * (n_blocks - 2))
    if n_blocks > 1:
        d_layers.append(d_last)

    def f0():
        dropouts = [0.1 * x for x in range(len(d_layers))]
        return rtdl.MLP(
            d_in=d, d_layers=d_layers, dropouts=dropouts, activation='GELU', d_out=d_out
        )

    def f1():
        return rtdl.MLP.make_baseline(
            d_in=d, d_layers=d_layers, dropout=0.1, d_out=d_out
        )

    model = locals()[f'f{constructor}']().to(device)
    n = 2
    assert model(torch.randn(n, d, device=device)).shape == (
        (n, d_out) if d_out else (n, d_last) if n_blocks > 1 else (n, d)
    )


@pytest.mark.parametrize('n_blocks', [1, 2])
@pytest.mark.parametrize('d_out', [1, 2])
@pytest.mark.parametrize('constructor', range(2))
@pytest.mark.parametrize('device', get_devices())
def test_resnet(n_blocks, d_out, constructor, device):
    d = 4

    def f0():
        return rtdl.ResNet.make_baseline(
            d_in=d,
            d_main=d,
            d_hidden=d * 3,
            dropout_first=0.1,
            dropout_second=0.2,
            n_blocks=n_blocks,
            d_out=d_out,
        )

    def f1():
        return rtdl.ResNet(
            d_in=d,
            d_main=d,
            d_hidden=d * 3,
            dropout_first=0.1,
            dropout_second=0.2,
            n_blocks=n_blocks,
            normalization='Identity',
            activation='ReLU6',
            d_out=d_out,
        )

    model = locals()[f'f{constructor}']().to(device)
    n = 2
    assert model(torch.randn(n, d, device=device)).shape == (
        (n, d_out) if d_out else (n, d)
    )


@pytest.mark.parametrize('n_blocks', range(1, 7))
@pytest.mark.parametrize('d_out', [1, 2])
@pytest.mark.parametrize('last_layer_query_idx', [None, [-1]])
@pytest.mark.parametrize('constructor', range(3))
@pytest.mark.parametrize('device', get_devices())
def test_ft_transformer(n_blocks, d_out, last_layer_query_idx, constructor, device):
    n_num_features = 4
    model = rtdl.FTTransformer.make_default(
        n_num_features=4,
        cat_cardinalities=[2, 3],
        n_blocks=n_blocks,
        last_layer_query_idx=last_layer_query_idx,
        kv_compression_ratio=0.5,
        kv_compression_sharing='headwise',
        d_out=d_out,
    ).to(device)
    n = 2

    # check that the following methods do not fail
    model.optimization_param_groups()
    model.make_default_optimizer()

    def f0():
        return rtdl.FTTransformer.make_default(
            n_num_features=4,
            cat_cardinalities=[2, 3],
            n_blocks=n_blocks,
            last_layer_query_idx=last_layer_query_idx,
            kv_compression_ratio=0.5,
            kv_compression_sharing='headwise',
            d_out=d_out,
        )

    def f1():
        return rtdl.FTTransformer.make_baseline(
            n_num_features=4,
            cat_cardinalities=[2, 3],
            n_blocks=n_blocks,
            d_token=8,
            attention_dropout=0.2,
            ffn_d_hidden=8,
            ffn_dropout=0.3,
            residual_dropout=0.4,
            last_layer_query_idx=last_layer_query_idx,
            kv_compression_ratio=0.5,
            kv_compression_sharing='headwise',
            d_out=d_out,
        )

    def f2():
        d_token = 8
        rtdl.Transformer.WARNINGS['prenormalization'] = False
        model = rtdl.FTTransformer(
            rtdl.FeatureTokenizer(4, [2, 3], d_token),
            rtdl.Transformer(
                d_token=d_token,
                attention_n_heads=1,
                attention_dropout=0.3,
                attention_initialization='xavier',
                attention_normalization='Identity',
                ffn_d_hidden=4,
                ffn_dropout=0.3,
                ffn_activation='SELU',
                residual_dropout=0.2,
                ffn_normalization='Identity',
                prenormalization=False,
                first_prenormalization=False,
                n_tokens=7,
                head_activation='PReLU',
                head_normalization='Identity',
                n_blocks=n_blocks,
                last_layer_query_idx=last_layer_query_idx,
                kv_compression_ratio=0.5,
                kv_compression_sharing='headwise',
                d_out=d_out,
            ),
        )
        rtdl.Transformer.WARNINGS['prenormalization'] = True
        return model

    model = locals()[f'f{constructor}']().to(device)
    x = model(
        torch.randn(n, n_num_features, device=device),
        torch.tensor([[0, 2], [1, 2]], device=device),
    )
    if d_out:
        assert x.shape == (n, d_out)
    else:
        assert x.shape == (
            n,
            model.feature_tokenizer.n_tokens + 1 if last_layer_query_idx is None else 1,
            model.feature_tokenizer.d_token,
        )



"""Tools for data (pre)processing."""

from typing import List

import numpy as np


def get_category_sizes(X: np.ndarray) -> List[int]:
    """Validate encoded categorical features and count distinct values.

    The function calculates the "category sizes" that can be used to construct
    `rtdl.CategoricalFeatureTokenizer` and `rtdl.FTTransformer`. Additionally, the
    following conditions are checked:

    * the data is a two-dimensional array of signed integers
    * distinct values of each column form zero-based ranges

    Note:
        For valid inputs, the result equals :code:`X.max(0) + 1`.

    Args:
        X: encoded categorical features (e.g. the output of :code:`sklearn.preprocessing.OrdinalEncoder`)

    Returns:
        The counts of distinct values for all columns.

    Examples:
        .. testcode::

            assert get_category_sizes(np.array(
                [
                    [0, 0, 0],
                    [1, 0, 0],
                    [2, 1, 0],
                ]
            )) == [3, 2, 1]
    """
    assert X.ndim == 2
    assert issubclass(X.dtype.type, np.signedinteger)
    sizes = []
    for i, column in enumerate(X.T):
        unique_values = np.unique(column)
        min_value = unique_values.min()
        assert (
            min_value == 0
        ), f'The minimum value of column {i} is {min_value}, but it must be zero.'
        max_value = unique_values.max()
        assert max_value + 1 == len(
            unique_values
        ), f'The values of column {i} do not fully cover the range from zero to maximum_value={max_value}'

        sizes.append(len(unique_values))
    return sizes



"""Research on Tabular Deep Learning."""

__version__ = '0.0.13'

from . import data  # noqa
from .functional import geglu, reglu  # noqa
from .modules import (  # noqa
    GEGLU,
    MLP,
    CategoricalFeatureTokenizer,
    CLSToken,
    FeatureTokenizer,
    FTTransformer,
    MultiheadAttention,
    NumericalFeatureTokenizer,
    ReGLU,
    ResNet,
    Transformer,
)



import torch.nn.functional as F
from torch import Tensor


def reglu(x: Tensor) -> Tensor:
    """The ReGLU activation function from [1].

    References:

        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.relu(b)


def geglu(x: Tensor) -> Tensor:
    """The GEGLU activation function from [1].

    References:

        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.gelu(b)



import numpy as np
import pytest

import rtdl.data


def test_get_category_sizes():
    get_category_sizes = rtdl.data.get_category_sizes

    # not two dimensions
    with pytest.raises(AssertionError):
        get_category_sizes(np.array([0, 0, 0]))
    with pytest.raises(AssertionError):
        get_category_sizes(np.array([[[0, 0, 0]]]))

    # not signed integers
    for dtype in [np.uint32, np.float32, str]:
        with pytest.raises(AssertionError):
            get_category_sizes(np.array([[0, 0, 0]], dtype=dtype))

    # non-zero min value
    for x in [-1, 1]:
        with pytest.raises(AssertionError):
            get_category_sizes(np.array([[0, 0, x]]))

    # not full range
    with pytest.raises(AssertionError):
        get_category_sizes(np.array([[0, 0, 0], [2, 1, 1]]))

    # correctness
    assert get_category_sizes(np.array([[0]])) == [1]
    assert get_category_sizes(np.array([[0], [1]])) == [2]
    assert get_category_sizes(np.array([[0, 0]])) == [1, 1]
    assert get_category_sizes(np.array([[0, 0], [0, 1]])) == [1, 2]
    assert get_category_sizes(np.array([[1, 0], [0, 1]])) == [2, 2]



# Tests in this file validate that the models in `rtdl` are LITERALLY THE SAME as
# the ones used in the paper (https://github.com/Yura52/rtdl/tree/main/bin)
# The testing approach:
# (1) copy weights from the correct model to the RTDL model
# (2) check that the two models produce the same output for the same input
import math
import random
import typing as ty

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from pytest import mark
from torch import Tensor

import rtdl


class Model(nn.Module):
    def __init__(self, cat_input_module: nn.Module, model: nn.Module):
        super().__init__()
        self.cat_input_module = cat_input_module
        self.model = model

    def forward(self, x_num, x_cat):
        return self.model(
            torch.cat([x_num, self.cat_input_module(x_cat).flatten(1, -1)], dim=1)
        )


def set_seeds(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


@torch.no_grad()
def copy_layer(dst: nn.Module, src: nn.Module):
    for key, _ in dst.named_parameters():
        getattr(dst, key).copy_(getattr(src, key))


@torch.no_grad()
@mark.parametrize('seed', range(10))
def test_mlp(seed):
    # Source: https://github.com/Yura52/rtdl/blob/0e5169659c7ce552bc05bbaa85f7e204adc3d88e/bin/mlp.py

    class CorrectMLP(nn.Module):
        def __init__(
            self,
            *,
            d_in: int,
            d_layers: ty.List[int],
            dropout: float,
            d_out: int,
            categories: ty.Optional[ty.List[int]],
            d_embedding: int,
        ) -> None:
            super().__init__()

            if categories is not None:
                d_in += len(categories) * d_embedding
                category_offsets = torch.tensor([0] + categories[:-1]).cumsum(0)
                self.register_buffer('category_offsets', category_offsets)
                self.category_embeddings = nn.Embedding(sum(categories), d_embedding)
                nn.init.kaiming_uniform_(
                    self.category_embeddings.weight, a=math.sqrt(5)
                )
                # print(f'{self.category_embeddings.weight.shape=}')

            self.layers = nn.ModuleList(
                [
                    nn.Linear(d_layers[i - 1] if i else d_in, x)
                    for i, x in enumerate(d_layers)
                ]
            )
            self.dropout = dropout
            self.head = nn.Linear(d_layers[-1] if d_layers else d_in, d_out)

        def forward(self, x_num, x_cat):
            if x_cat is not None:
                x_cat = self.category_embeddings(x_cat + self.category_offsets[None])  # type: ignore
                x = torch.cat([x_num, x_cat.view(x_cat.size(0), -1)], dim=-1)
            else:
                x = x_num

            for layer in self.layers:
                x = layer(x)
                x = F.relu(x)
                if self.dropout:
                    x = F.dropout(x, self.dropout, self.training)
            x = self.head(x)
            return x

    n = 32
    d_num = 2
    categories = [2, 3]
    d_embedding = 3
    d_in = d_num + len(categories) * d_embedding
    d_layers = [3, 4, 5]
    dropout = 0.1
    d_out = 2

    set_seeds(seed)
    correct_model = CorrectMLP(
        d_in=d_num,
        d_layers=d_layers,
        dropout=dropout,
        d_out=d_out,
        categories=categories,
        d_embedding=d_embedding,
    )
    rtdl_tokenizer = rtdl.CategoricalFeatureTokenizer(
        categories, d_embedding, False, 'uniform'
    )
    rtdl_backbone = rtdl.MLP(
        d_in=d_in,
        d_layers=d_layers,
        dropouts=dropout,
        activation='ReLU',
        d_out=d_out,
    )

    rtdl_tokenizer.embeddings.weight.copy_(correct_model.category_embeddings.weight)
    for correct_layer, block in zip(correct_model.layers, rtdl_backbone.blocks):
        copy_layer(block.linear, correct_layer)
    copy_layer(rtdl_backbone.head, correct_model.head)

    rtdl_model = Model(rtdl_tokenizer, rtdl_backbone)
    x_num = torch.randn(n, d_num)
    x_cat = torch.cat([torch.randint(x, (n, 1)) for x in categories], dim=1)
    set_seeds(seed)
    correct_result = correct_model(x_num, x_cat)
    set_seeds(seed)
    rtdl_result = rtdl_model(x_num, x_cat)
    assert (correct_result == rtdl_result).all()


@torch.no_grad()
@mark.parametrize('seed', range(10))
def test_resnet(seed):
    # Source: https://github.com/Yura52/rtdl/blob/0e5169659c7ce552bc05bbaa85f7e204adc3d88e/bin/resnet.py

    class CorrectResNet(nn.Module):
        def __init__(
            self,
            *,
            d_numerical: int,
            categories: ty.Optional[ty.List[int]],
            d_embedding: int,
            d: int,
            d_hidden_factor: float,
            n_layers: int,
            activation: str,
            normalization: str,
            hidden_dropout: float,
            residual_dropout: float,
            d_out: int,
        ) -> None:
            super().__init__()

            def make_normalization():
                return {'BatchNorm1d': nn.BatchNorm1d}[normalization](d)

            assert activation == 'ReLU'
            self.main_activation = F.relu
            self.last_activation = F.relu
            self.residual_dropout = residual_dropout
            self.hidden_dropout = hidden_dropout

            d_in = d_numerical
            d_hidden = int(d * d_hidden_factor)

            if categories is not None:
                d_in += len(categories) * d_embedding
                category_offsets = torch.tensor([0] + categories[:-1]).cumsum(0)
                self.register_buffer('category_offsets', category_offsets)
                self.category_embeddings = nn.Embedding(sum(categories), d_embedding)
                nn.init.kaiming_uniform_(
                    self.category_embeddings.weight, a=math.sqrt(5)
                )
                # print(f'{self.category_embeddings.weight.shape=}')

            self.first_layer = nn.Linear(d_in, d)
            self.layers = nn.ModuleList(
                [
                    nn.ModuleDict(
                        {
                            'norm': make_normalization(),
                            'linear0': nn.Linear(
                                d, d_hidden * (2 if activation.endswith('glu') else 1)
                            ),
                            'linear1': nn.Linear(d_hidden, d),
                        }
                    )
                    for _ in range(n_layers)
                ]
            )
            self.last_normalization = make_normalization()
            self.head = nn.Linear(d, d_out)

        def forward(self, x_num: Tensor, x_cat: ty.Optional[Tensor]) -> Tensor:
            if x_cat is not None:
                x_cat = self.category_embeddings(x_cat + self.category_offsets[None])  # type: ignore
                x = torch.cat([x_num, x_cat.view(x_cat.size(0), -1)], dim=-1)  # type: ignore
            else:
                x = x_num

            x = self.first_layer(x)
            for layer in self.layers:
                layer = ty.cast(ty.Dict[str, nn.Module], layer)
                z = x
                z = layer['norm'](z)
                z = layer['linear0'](z)
                z = self.main_activation(z)
                if self.hidden_dropout:
                    z = F.dropout(z, self.hidden_dropout, self.training)
                z = layer['linear1'](z)
                if self.residual_dropout:
                    z = F.dropout(z, self.residual_dropout, self.training)
                x = x + z
            x = self.last_normalization(x)
            x = self.last_activation(x)
            x = self.head(x)
            return x

    n = 32
    d_num = 2
    categories = [2, 3]
    d_embedding = 3
    d_in = d_num + len(categories) * d_embedding
    d = 4
    d_hidden_factor = 1.5
    n_layers = 2
    activation = 'ReLU'
    normalization = 'BatchNorm1d'
    hidden_dropout = 0.1
    residual_dropout = 0.2
    d_out = 2

    set_seeds(seed)
    correct_model = CorrectResNet(
        d_numerical=d_num,
        categories=categories,
        d_embedding=d_embedding,
        d=d,
        d_hidden_factor=d_hidden_factor,
        n_layers=n_layers,
        activation=activation,
        normalization=normalization,
        hidden_dropout=0.1,
        residual_dropout=0.2,
        d_out=d_out,
    )
    rtdl_tokenizer = rtdl.CategoricalFeatureTokenizer(
        categories, d_embedding, False, 'uniform'
    )
    rtdl_backbone = rtdl.ResNet(
        d_in=d_in,
        n_blocks=n_layers,
        d_main=d,
        d_hidden=int(d * d_hidden_factor),
        dropout_first=hidden_dropout,
        dropout_second=residual_dropout,
        normalization=normalization,
        activation=activation,
        d_out=d_out,
    )

    rtdl_tokenizer.embeddings.weight.copy_(correct_model.category_embeddings.weight)
    copy_layer(rtdl_backbone.first_layer, correct_model.first_layer)
    for correct_layer, block in zip(correct_model.layers, rtdl_backbone.blocks):
        copy_layer(block.normalization, correct_layer['norm'])
        copy_layer(block.linear_first, correct_layer['linear0'])
        copy_layer(block.linear_second, correct_layer['linear1'])

    copy_layer(rtdl_backbone.head.normalization, correct_model.last_normalization)
    copy_layer(rtdl_backbone.head.linear, correct_model.head)

    rtdl_model = Model(rtdl_tokenizer, rtdl_backbone)
    x_num = torch.randn(n, d_num)
    x_cat = torch.cat([torch.randint(x, (n, 1)) for x in categories], dim=1)
    set_seeds(seed)
    correct_result = correct_model(x_num, x_cat)
    set_seeds(seed)
    rtdl_result = rtdl_model(x_num, x_cat)
    assert (correct_result == rtdl_result).all()


@torch.no_grad()
@mark.parametrize('seed', range(10))
@mark.parametrize('kv_compression_ratio', [None, 0.5])
def test_ft_transformer(seed, kv_compression_ratio):
    # """Source: https://github.com/Yura52/rtdl/blob/0e5169659c7ce552bc05bbaa85f7e204adc3d88e/bin/ft_transformer.py"""
    # The only difference is that [CLS] is now the last token.

    def correct_reglu(x: Tensor) -> Tensor:
        a, b = x.chunk(2, dim=-1)
        return a * F.relu(b)

    class CorrectTokenizer(nn.Module):
        category_offsets: ty.Optional[Tensor]

        def __init__(
            self,
            d_numerical: int,
            categories: ty.Optional[ty.List[int]],
            d_token: int,
            bias: bool,
        ) -> None:
            super().__init__()
            if categories is None:
                d_bias = d_numerical
                self.category_offsets = None
                self.category_embeddings = None
            else:
                d_bias = d_numerical + len(categories)
                category_offsets = torch.tensor([0] + categories[:-1]).cumsum(0)
                self.register_buffer('category_offsets', category_offsets)
                self.category_embeddings = nn.Embedding(sum(categories), d_token)
                nn.init.kaiming_uniform_(
                    self.category_embeddings.weight, a=math.sqrt(5)
                )

            # take [CLS] token into account
            self.weight = nn.Parameter(Tensor(d_numerical + 1, d_token))
            self.bias = nn.Parameter(Tensor(d_bias, d_token)) if bias else None
            # The initialization is inspired by nn.Linear
            nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
            if self.bias is not None:
                nn.init.kaiming_uniform_(self.bias, a=math.sqrt(5))

        @property
        def n_tokens(self) -> int:
            return len(self.weight) + (
                0 if self.category_offsets is None else len(self.category_offsets)
            )

        def forward(self, x_num: Tensor, x_cat: ty.Optional[Tensor]) -> Tensor:
            # x_num = torch.cat(
            #     [
            #         torch.ones(len(x_num), 1, device=x_num.device),
            #         x_num,
            #     ],
            #     dim=1,
            # )
            x = self.weight[:-1][None] * x_num[:, :, None]
            if x_cat is not None:
                x = torch.cat(
                    [x, self.category_embeddings(x_cat + self.category_offsets[None])],
                    dim=1,
                )
            x = torch.cat(
                [x, self.weight[-1][None, None].repeat(len(x), 1, 1)],
                dim=1,
            )
            if self.bias is not None:
                bias = torch.cat(
                    [
                        self.bias,
                        torch.zeros(1, self.bias.shape[1], device=x_num.device),
                    ]
                )
                x = x + bias[None]
            return x

    class CorrectMultiheadAttention(nn.Module):
        def __init__(
            self, d: int, n_heads: int, dropout: float, initialization: str
        ) -> None:
            if n_heads > 1:
                assert d % n_heads == 0
            assert initialization in ['xavier', 'kaiming']

            super().__init__()
            self.W_q = nn.Linear(d, d)
            self.W_k = nn.Linear(d, d)
            self.W_v = nn.Linear(d, d)
            self.W_out = nn.Linear(d, d) if n_heads > 1 else None
            self.n_heads = n_heads
            self.dropout = nn.Dropout(dropout) if dropout else None

            for m in [self.W_q, self.W_k, self.W_v]:
                if initialization == 'xavier' and (n_heads > 1 or m is not self.W_v):
                    # gain is needed since W_qkv is represented with 3 separate layers
                    nn.init.xavier_uniform_(m.weight, gain=1 / math.sqrt(2))
                nn.init.zeros_(m.bias)
            if self.W_out is not None:
                nn.init.zeros_(self.W_out.bias)

        def _reshape(self, x: Tensor) -> Tensor:
            batch_size, n_tokens, d = x.shape
            d_head = d // self.n_heads
            return (
                x.reshape(batch_size, n_tokens, self.n_heads, d_head)
                .transpose(1, 2)
                .reshape(batch_size * self.n_heads, n_tokens, d_head)
            )

        def forward(
            self,
            x_q: Tensor,
            x_kv: Tensor,
            key_compression: ty.Optional[nn.Linear],
            value_compression: ty.Optional[nn.Linear],
        ) -> Tensor:
            q, k, v = self.W_q(x_q), self.W_k(x_kv), self.W_v(x_kv)
            for tensor in [q, k, v]:
                assert tensor.shape[-1] % self.n_heads == 0
            if key_compression is not None:
                assert value_compression is not None
                k = key_compression(k.transpose(1, 2)).transpose(1, 2)
                v = value_compression(v.transpose(1, 2)).transpose(1, 2)
            else:
                assert value_compression is None

            batch_size = len(q)
            d_head_key = k.shape[-1] // self.n_heads
            d_head_value = v.shape[-1] // self.n_heads
            n_q_tokens = q.shape[1]

            q = self._reshape(q)
            k = self._reshape(k)
            attention = F.softmax(q @ k.transpose(1, 2) / math.sqrt(d_head_key), dim=-1)
            if self.dropout is not None:
                attention = self.dropout(attention)
            x = attention @ self._reshape(v)
            x = (
                x.reshape(batch_size, self.n_heads, n_q_tokens, d_head_value)
                .transpose(1, 2)
                .reshape(batch_size, n_q_tokens, self.n_heads * d_head_value)
            )
            if self.W_out is not None:
                x = self.W_out(x)
            return x

    class CorrectFTTransformer(nn.Module):
        def __init__(
            self,
            *,
            # tokenizer
            d_numerical: int,
            categories: ty.Optional[ty.List[int]],
            token_bias: bool,
            # transformer
            n_layers: int,
            d_token: int,
            n_heads: int,
            d_ffn_factor: float,
            attention_dropout: float,
            ffn_dropout: float,
            residual_dropout: float,
            activation: str,
            prenormalization: bool,
            initialization: str,
            # linformer
            kv_compression: ty.Optional[float],
            kv_compression_sharing: ty.Optional[str],
            #
            d_out: int,
        ) -> None:
            assert (kv_compression is None) ^ (kv_compression_sharing is not None)

            super().__init__()
            self.tokenizer = CorrectTokenizer(
                d_numerical, categories, d_token, token_bias
            )
            n_tokens = self.tokenizer.n_tokens

            def make_kv_compression():
                assert kv_compression
                compression = nn.Linear(
                    n_tokens, int(n_tokens * kv_compression), bias=False
                )
                if initialization == 'xavier':
                    nn.init.xavier_uniform_(compression.weight)
                return compression

            self.shared_kv_compression = (
                make_kv_compression()
                if kv_compression and kv_compression_sharing == 'layerwise'
                else None
            )

            def make_normalization():
                return nn.LayerNorm(d_token)

            d_hidden = int(d_token * d_ffn_factor)
            self.layers = nn.ModuleList([])
            for layer_idx in range(n_layers):
                layer = nn.ModuleDict(
                    {
                        'attention': CorrectMultiheadAttention(
                            d_token, n_heads, attention_dropout, initialization
                        ),
                        'linear0': nn.Linear(
                            d_token, d_hidden * (2 if activation.endswith('glu') else 1)
                        ),
                        'linear1': nn.Linear(d_hidden, d_token),
                        'norm1': make_normalization(),
                    }
                )
                if not prenormalization or layer_idx:
                    layer['norm0'] = make_normalization()
                if kv_compression and self.shared_kv_compression is None:
                    layer['key_compression'] = make_kv_compression()
                    if kv_compression_sharing == 'headwise':
                        layer['value_compression'] = make_kv_compression()
                    else:
                        assert kv_compression_sharing == 'key-value'
                self.layers.append(layer)

            assert activation == 'reglu'
            self.activation = correct_reglu
            self.last_activation = F.relu
            self.prenormalization = prenormalization
            self.last_normalization = make_normalization() if prenormalization else None
            self.ffn_dropout = ffn_dropout
            self.residual_dropout = residual_dropout
            self.head = nn.Linear(d_token, d_out)

        def _get_kv_compressions(self, layer):
            return (
                (self.shared_kv_compression, self.shared_kv_compression)
                if self.shared_kv_compression is not None
                else (layer['key_compression'], layer['value_compression'])
                if 'key_compression' in layer and 'value_compression' in layer
                else (layer['key_compression'], layer['key_compression'])
                if 'key_compression' in layer
                else (None, None)
            )

        def _start_residual(self, x, layer, norm_idx):
            x_residual = x
            if self.prenormalization:
                norm_key = f'norm{norm_idx}'
                if norm_key in layer:
                    x_residual = layer[norm_key](x_residual)
            return x_residual

        def _end_residual(self, x, x_residual, layer, norm_idx):
            if self.residual_dropout:
                x_residual = F.dropout(x_residual, self.residual_dropout, self.training)
            x = x + x_residual
            if not self.prenormalization:
                x = layer[f'norm{norm_idx}'](x)
            return x

        def forward(self, x_num: Tensor, x_cat: ty.Optional[Tensor]) -> Tensor:
            x = self.tokenizer(x_num, x_cat)

            for layer_idx, layer in enumerate(self.layers):
                is_last_layer = layer_idx + 1 == len(self.layers)
                layer = ty.cast(ty.Dict[str, nn.Module], layer)

                x_residual = self._start_residual(x, layer, 0)
                x_residual = layer['attention'](
                    # for the last attention, it is enough to process only [CLS]
                    (x_residual[:, -1:] if is_last_layer else x_residual),
                    x_residual,
                    *self._get_kv_compressions(layer),
                )
                if is_last_layer:
                    x = x[:, -1:]
                x = self._end_residual(x, x_residual, layer, 0)

                x_residual = self._start_residual(x, layer, 1)
                x_residual = layer['linear0'](x_residual)
                x_residual = self.activation(x_residual)
                if self.ffn_dropout:
                    x_residual = F.dropout(x_residual, self.ffn_dropout, self.training)
                x_residual = layer['linear1'](x_residual)
                x = self._end_residual(x, x_residual, layer, 1)

            assert x.shape[1] == 1
            x = x[:, 0]
            if self.last_normalization is not None:
                x = self.last_normalization(x)
            x = self.last_activation(x)
            x = self.head(x)
            x = x.squeeze(-1)
            return x

    # Source: https://github.com/Yura52/rtdl/blob/0e5169659c7ce552bc05bbaa85f7e204adc3d88e/output/california_housing/ft_transformer/default/0.toml
    default_config = {
        'seed': 0,
        'data': {
            'normalization': 'quantile_normal',
            'path': 'data/california_housing',
            'y_policy': 'mean_std',
        },
        'model': {
            'activation': 'reglu',
            'attention_dropout': 0.2,
            'd_ffn_factor': 4 / 3,
            'd_token': 192,
            'ffn_dropout': 0.1,
            'initialization': 'kaiming',
            'n_heads': 8,
            'n_layers': 3,
            'prenormalization': True,
            'residual_dropout': 0.0,
        },
        'training': {
            'batch_size': 256,
            'eval_batch_size': 8192,
            'lr': 0.0001,
            'lr_n_decays': 0,
            'n_epochs': 1000000000,
            'optimizer': 'adamw',
            'patience': 16,
            'weight_decay': 1e-05,
        },
    }
    dcm = default_config['model']

    n = 4
    d_num = 2
    categories = [2, 3]
    n_tokens = d_num + len(categories) + 1
    kv_compression_sharing = 'key-value' if kv_compression_ratio else None
    d_out = 2

    set_seeds(seed)
    correct_model = CorrectFTTransformer(
        d_numerical=d_num,
        categories=categories,
        token_bias=True,
        kv_compression=kv_compression_ratio,
        kv_compression_sharing=kv_compression_sharing,
        d_out=d_out,
        **dcm,
    )
    rtdl_model = rtdl.FTTransformer(
        rtdl.FeatureTokenizer(d_num, categories, dcm['d_token']),
        rtdl.Transformer(
            d_token=dcm['d_token'],
            n_blocks=dcm['n_layers'],
            attention_n_heads=dcm['n_heads'],
            attention_dropout=dcm['attention_dropout'],
            attention_initialization=dcm['initialization'],
            attention_normalization='LayerNorm',
            ffn_d_hidden=int(dcm['d_token'] * dcm['d_ffn_factor']),
            ffn_dropout=dcm['ffn_dropout'],
            ffn_activation='ReGLU',
            ffn_normalization='LayerNorm',
            residual_dropout=dcm['residual_dropout'],
            prenormalization=dcm['prenormalization'],
            first_prenormalization=False,
            last_layer_query_idx=[-1],
            n_tokens=n_tokens if kv_compression_ratio else None,
            kv_compression_ratio=kv_compression_ratio,
            kv_compression_sharing=kv_compression_sharing,
            head_activation='ReLU',
            head_normalization='LayerNorm',
            d_out=d_out,
        ),
    )
    rtdl_default_model = rtdl.FTTransformer.make_default(
        n_num_features=d_num,
        cat_cardinalities=categories,
        last_layer_query_idx=[-1],
        kv_compression_ratio=kv_compression_ratio,
        kv_compression_sharing=kv_compression_sharing,
        d_out=d_out,
    )

    rtdl_model.feature_tokenizer.num_tokenizer.weight.copy_(
        correct_model.tokenizer.weight[:-1]
    )
    rtdl_model.feature_tokenizer.num_tokenizer.bias.copy_(
        correct_model.tokenizer.bias[:d_num]
    )
    rtdl_model.feature_tokenizer.cat_tokenizer.embeddings.weight.copy_(
        correct_model.tokenizer.category_embeddings.weight
    )
    rtdl_model.feature_tokenizer.cat_tokenizer.bias.copy_(
        correct_model.tokenizer.bias[-len(categories) :]
    )
    rtdl_model.cls_token.weight.copy_(correct_model.tokenizer.weight[-1])
    for correct_layer, block in zip(
        correct_model.layers, rtdl_model.transformer.blocks
    ):
        for key in ['W_q', 'W_k', 'W_v', 'W_out']:
            copy_layer(
                getattr(block['attention'], key),
                getattr(correct_layer['attention'], key),
            )
        copy_layer(block['ffn'].linear_first, correct_layer['linear0'])
        copy_layer(block['ffn'].linear_second, correct_layer['linear1'])
        copy_layer(block['ffn_normalization'], correct_layer['norm1'])
        if 'norm0' in correct_layer:
            copy_layer(block['attention_normalization'], correct_layer['norm0'])
        for key in ['key_compression', 'value_compression']:
            if key in correct_layer:
                copy_layer(block[key], correct_layer[key])
    copy_layer(
        rtdl_model.transformer.head.normalization, correct_model.last_normalization
    )
    copy_layer(rtdl_model.transformer.head.linear, correct_model.head)
    rtdl_default_model.load_state_dict(rtdl_model.state_dict())

    x_num = torch.randn(n, d_num)
    x_cat = torch.cat([torch.randint(x, (n, 1)) for x in categories], dim=1)
    results = []
    for m in [correct_model, rtdl_model, rtdl_default_model]:
        set_seeds(seed)
        results.append(m(x_num, x_cat))
    correct_result = results[0]
    assert (results[1] == results[2]).all()
    assert (results[1] == correct_result).all()
